@@ -3,27 +3,26 @@
 
 /*---
 esid: sec-array.prototype.reduceright
-es5id: 15.4.4.22-8-b-iii-1-18
+es5id: 15.4.4.22-8-b-iii-1-6
 description: >
-    Array.prototype.reduceRight - element to be retrieved is own
-    accessor property without a get function on an Array
+    Array.prototype.reduceRight - element to be retrieved is own data
+    property that overrides an inherited accessor property on an Array
 ---*/
 
 var testResult = false;
 
 function callbackfn(prevVal, curVal, idx, obj) {
   if (idx === 1) {
-    testResult = (typeof prevVal === "undefined");
+    testResult = (prevVal === 2);
   }
 }
 
-var arr = [0, 1];
-
-Object.defineProperty(arr, "2", {
-  set: function() {},
+Object.defineProperty(Array.prototype, "2", {
+  get: function() {
+    return "2";
+  },
   configurable: true
 });
-
-arr.reduceRight(callbackfn);
+[0, 1, 2].reduceRight(callbackfn);
 
 assert(testResult, 'testResult !== true');
