@@ -137,6 +137,8 @@ private slots:
     void revertListBug();
     void QTBUG_38492();
     void revertListMemoryLeak();
+    void duplicateStateName();
+    void trivialWhen();
 };
 
 void tst_qquickstates::initTestCase()
@@ -1654,6 +1656,26 @@ void tst_qquickstates::revertListMemoryLeak()
     QVERIFY(bindingPtr->ref == 1);
 }
 
+void tst_qquickstates::duplicateStateName()
+{
+    QQmlEngine engine;
+
+    QQmlComponent c(&engine, testFileUrl("duplicateStateName.qml"));
+    QTest::ignoreMessage(QtWarningMsg, fullDataPath("duplicateStateName.qml") + ":3:1: QML Rectangle: Found duplicate state name: state1");
+    QScopedPointer<QQuickItem> item(qobject_cast<QQuickItem *>(c.create()));
+    QVERIFY(!item.isNull());
+}
+
+// QTBUG-76838
+void tst_qquickstates::trivialWhen()
+{
+    QQmlEngine engine;
+
+    QQmlComponent c(&engine, testFileUrl("trivialWhen.qml"));
+    QVERIFY(c.create());
+}
+
+
 QTEST_MAIN(tst_qquickstates)
 
 #include "tst_qquickstates.moc"
